@@ -93,14 +93,14 @@ runTest settings damage = withTmpDir "tmprepo" $ \tmpdir -> do
 		]
 	unless cloned $
 		error $ "failed to clone this repo"
-	g <- Git.Config.read =<< Git.Construct.fromPath cloneloc
+	g <- Git.Config.read =<< Git.Construct.fromPath (toRawFilePath cloneloc)
 	Git.Destroyer.applyDamage damage g
 	repairstatus <- catchMaybeIO $ Git.Repair.successfulRepair
 		<$> Git.Repair.runRepair Git.Repair.isTrackingBranch (forced settings) g
 	case repairstatus of
 		Just True -> testResult repairstatus 
 			. Just . not . Git.Fsck.foundBroken
-			=<< Git.Fsck.findBroken False g
+			=<< Git.Fsck.findBroken False False g
 		_ -> testResult repairstatus Nothing
 
 -- Pass test result and fsck result
